#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections.abc import Collection, Container
from pathlib import Path
from typing import Any, Callable, Union

from ycontract import ContractError

__all__ = [
    'more_than', 'less_than', 'more_than_or_eq', 'less_than_or_eq', 'in_range', 'exists_path'
]


def _to_native_number(x) -> float:
    if hasattr(x, '__float__'):
        return float(x)
    elif hasattr(x, '__int__'):
        return int(x)
    else:
        raise ContractError(f"{x} is not a number.")


def more_than(more_than_target) -> Callable[[Any], bool]:
    more_than_target = _to_native_number(more_than_target)

    def _more_than(x) -> bool:
        return _to_native_number(x) > more_than_target

    return _more_than


is_positive = more_than(0.0)


def less_than(less_than_target) -> Callable[[Any], bool]:
    less_than_target = _to_native_number(less_than_target)

    def _less_than(x) -> bool:
        return _to_native_number(x) < less_than_target

    return _less_than


is_negative = less_than(0.0)


def more_than_or_eq(more_than_target) -> Callable[[Any], bool]:
    more_than_target = _to_native_number(more_than_target)

    def _more_than_or_eq(x) -> bool:
        return _to_native_number(x) >= more_than_target

    return _more_than_or_eq


def less_than_or_eq(less_than_target) -> Callable[[Any], bool]:
    less_than_target = _to_native_number(less_than_target)

    def _less_than_or_eq(x) -> bool:
        return _to_native_number(x) <= less_than_target

    return _less_than_or_eq


def in_range(a, b=None) -> Callable[[Any], bool]:
    if b is None:
        start, until = 0, a
    else:
        start, until = a, b

    def _in_range(x) -> bool:
        return start <= x < until

    return _in_range


def is_sub(sup: Collection) -> Callable[[Any], bool]:
    set_sup = set(sup)

    def _is_sub(sub: Collection):
        return set(sub) <= set_sup

    return _is_sub


def is_type(type_: type) -> Callable[[Any], bool]:

    def _is_type(x) -> bool:
        return type(x) is type_

    return _is_type


def is_instance(type_: type) -> Callable[[Any], bool]:

    def _is_instance(x) -> bool:
        return isinstance(x, type_)

    return _is_instance


def is_in(c: Container) -> Callable[[Any], bool]:

    def _is_in(x) -> bool:
        return x in c

    return _is_in


def exists_path(p: Union[str, Path]) -> bool:
    return Path(p).exists()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
from copy import deepcopy
from inspect import signature
from typing import Callable, Dict, List, Optional

from ._in_contract import in_contract
from ._out_contract import out_contract
from .ycontract import DEFAULT_CONTRACT_STATE, ContractState

__all__ = ['contract']


def pops_out_cond_f(cond_kw: Dict[str, Callable], f: Callable) -> List[Callable[..., bool]]:
    sig = signature(f)
    f_varnames = set(sig.parameters.keys())
    usable_varnames = {'returns', 'results', 'result', 'res'}
    for f_varname in f_varnames:
        usable_varnames.discard(f_varname)

    cond_fs = []
    for usable_varname in usable_varnames:
        cond = cond_kw.pop(usable_varname, None)
        if cond is not None:
            if callable(cond):
                cond_fs.append(cond)
            elif hasattr(cond, '__iter__'):
                cond_fs.extend(cond)
    return cond_fs


def contract(
        *conds,
        contract_tag: Optional[str] = None,
        contract_state: ContractState = DEFAULT_CONTRACT_STATE,
        **cond_opts_) -> Callable:

    def _contract(f: Callable) -> Callable:
        if contract_state.is_disable:
            return f

        cond_opts = deepcopy(cond_opts_)
        cond_fs = pops_out_cond_f(cond_opts, f)

        @in_contract(*conds, contract_tag=contract_tag, contract_state=contract_state, **cond_opts)
        @out_contract(*cond_fs, contract_tag=contract_tag, contract_state=contract_state)
        @functools.wraps(f)
        def wrapped(*args, **kw) -> Callable:
            return f(*args, **kw)

        return wrapped

    return _contract



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dataclasses import dataclass

__all__ = [
    'DEFAULT_CONTRACT_STATE', 'ContractError', 'ContractException', 'ContractState',
    'disable_contract'
]


class ContractError(Exception):
    ...


class ContractException(Exception):
    ...


@dataclass
class ContractState:
    __is_enable: bool = True

    def enable(self) -> None:
        self.__is_enable = True

    def disable(self) -> None:
        self.__is_enable = False

    @property
    def is_enable(self) -> bool:
        return self.__is_enable

    @property
    def is_disable(self) -> bool:
        return not self.__is_enable

    def __str__(self) -> str:
        return f"ContractState<is_enable={self.__is_enable}>"

    def __repr__(self) -> str:
        return str(self)


DEFAULT_CONTRACT_STATE = ContractState()


def disable_contract(state: ContractState = DEFAULT_CONTRACT_STATE) -> None:
    state.disable()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
from inspect import getcallargs, signature
from typing import Any, Callable, Dict, Optional, Sequence, Union

from .ycontract import DEFAULT_CONTRACT_STATE, ContractError, ContractException, ContractState

__all__ = ['InContractException', 'in_contract']

VARINFO = Union[str, Sequence[str]]
CONDINFO = Union[Callable[..., bool], Sequence[Callable[..., bool]]]


class InContractException(ContractException):
    ...


def new_in_contract_exception(
        f: Callable, f_args: Sequence, f_kw: dict, cond_f: Callable[..., bool],
        contract_tag: Optional[str]) -> InContractException:
    callargs = getcallargs(f, *f_args, **f_kw)

    f_filename = f.__code__.co_filename
    f_lineno = f.__code__.co_firstlineno
    cond_f_filename = cond_f.__code__.co_filename
    cond_f_lineno = cond_f.__code__.co_firstlineno

    msg = "\n"
    if contract_tag:
        msg += f"    tag:       {contract_tag}\n"
    msg += f"    function:  {f.__name__} at {f_filename}:{f_lineno}\n"
    msg += f"    condition: {cond_f.__name__} at {cond_f_filename}:{cond_f_lineno}\n"
    arguments_str = ", ".join(f"{key}={value}" for key, value in callargs.items())
    msg += f"    arguments: {arguments_str}"

    return InContractException(msg)


def in_contract_callable_full_match_case(
        cond_f: Callable[..., bool], f: Callable, contract_tag: Optional[str],
        contract_state: ContractState, *args, **kw) -> bool:
    new_values = list(getcallargs(f, *args, **kw).values())
    return cond_f(*new_values)


def in_contract_callable_partial_match_case(
        cond_f: Callable[..., bool], f: Callable, contract_tag: Optional[str],
        contract_state: ContractState, *args, **kw) -> bool:
    f_sig = signature(f)
    f_varnames = list(f_sig.parameters)
    cond_sig = signature(cond_f)
    cond_varnames = list(cond_sig.parameters)
    varname_indexes = {v: f_varnames.index(v) for v in cond_varnames}

    def get_arg(v: str):
        if varname_indexes[v] < len(args):
            return args[varname_indexes[v]]
        elif v in kw:
            return kw[v]
        else:
            return f_sig.parameters[v].default

    var_args = {v: get_arg(v) for v in cond_sig.parameters.keys()}

    return cond_f(**var_args)


def in_contract_callable_one(
        cond_f: Callable[..., bool], f: Callable, contract_tag: Optional[str],
        contract_state: ContractState, *args, **kw) -> bool:
    try:
        return in_contract_callable_full_match_case(
            cond_f, f, contract_tag, contract_state, *args, **kw)
    except TypeError:
        return in_contract_callable_partial_match_case(
            cond_f, f, contract_tag, contract_state, *args, **kw)


def check_in_contract_dict(
        cond: Dict[VARINFO, Callable[..., bool]], f: Callable, contract_tag: Optional[str],
        contract_state: ContractState, args: Sequence,
        kw: Dict[str, Any]) -> Optional[InContractException]:
    for varinfo, cond_info in cond.items():
        cond_fs = []
        if callable(cond_info):
            cond_fs = [cond_info]
        else:
            cond_fs = list(cond_info)

        for cond_f in cond_fs:
            ex = get_contract_exception_from_dict(
                cond_f, f, varinfo, contract_tag, contract_state, *args, **kw)
            if ex is not None:
                return ex

    return None


def in_contract_dict_match_varname_case(
        varname: str, cond_f: Callable[..., bool], f: Callable, contract_tag: Optional[str],
        contract_state: ContractState, *args, **kw):
    sig = signature(f)
    binded = sig.bind(*args, **kw)
    binded.apply_defaults()
    keys_ = list(binded.arguments.keys())
    values = list(binded.arguments.values())
    ind = keys_.index(varname)
    is_ok = cond_f(values[ind])
    if is_ok:
        return True
    else:
        return values[ind]


def get_contract_exception_from_dict(
        cond_f: Callable[..., bool], f: Callable, varinfo, contract_tag: Optional[str],
        contract_state: ContractState, *args, **kw) -> Optional[InContractException]:
    if isinstance(varinfo, str):
        res = in_contract_dict_match_varname_case(
            varinfo, cond_f, f, contract_tag, contract_state, *args, **kw)
        if res is not True:
            return new_in_contract_exception(f, list(args), kw, cond_f, contract_tag)
    else:
        f_sig = signature(f)
        binded = f_sig.bind(*args, **kw)
        binded.apply_defaults()
        keys_ = list(binded.arguments.keys())
        values = list(binded.arguments.values())
        inds = []
        for varname in varinfo:
            inds.append(keys_.index(varname))
        cond_arguments = [values[ind] for ind in inds]
        if not cond_f(*cond_arguments):
            return new_in_contract_exception(f, list(args), kw, cond_f, contract_tag)

    return None


def in_contract(
        *conds,
        contract_tag: Optional[str] = None,
        contract_state: ContractState = DEFAULT_CONTRACT_STATE,
        **cond_opts) -> Callable:

    def _in_contract(f: Callable) -> Callable:

        @functools.wraps(f)
        def wrapped(*args, **kw) -> Callable:
            if contract_state.is_disable:
                return f(*args, **kw)

            try:
                for cond in conds:
                    if callable(cond):
                        if not in_contract_callable_one(cond, f, contract_tag, contract_state, *
                                                        args, **kw):
                            raise new_in_contract_exception(f, list(args), kw, cond, contract_tag)
                    else:
                        ex = check_in_contract_dict(
                            cond, f, contract_tag, contract_state, args, kw)
                        if ex is not None:
                            raise ex

                ex = check_in_contract_dict(
                    cond_opts, f, contract_tag, contract_state, args, kw)  # type: ignore
                if ex is not None:
                    raise ex
            except (ValueError, TypeError) as err:
                raise ContractError(*err.args)

            return f(*args, **kw)

        return wrapped

    return _in_contract



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import functools
from inspect import getcallargs, signature
from typing import Any, Callable, Dict, Optional, TypeVar

from .ycontract import DEFAULT_CONTRACT_STATE, ContractError, ContractException, ContractState

__all__ = ['OutContractException', 'out_contract']


class OutContractException(ContractException):
    ...


T = TypeVar('T')


def new_out_contract_exception(
        f: Callable[..., T], f_args: list, f_kw: dict, cond_f: Callable[..., bool], result: T,
        contract_tag: Optional[str]) -> OutContractException:
    callargs = getcallargs(f, *f_args, **f_kw)

    f_filename = f.__code__.co_filename
    f_lineno = f.__code__.co_firstlineno
    cond_f_filename = cond_f.__code__.co_filename
    cond_f_lineno = cond_f.__code__.co_firstlineno

    msg = "\n"
    if contract_tag:
        msg += f"    tag:       {contract_tag}\n"
    msg += f"    function:  {f.__name__} at {f_filename}:{f_lineno}\n"
    msg += f"    condition: {cond_f.__name__} at {cond_f_filename}:{cond_f_lineno}\n"
    arguments_str = ", ".join(f"{key}={value}" for key, value in callargs.items())
    msg += f"    arguments: {arguments_str}\n"
    msg += f"    result:    {result}"

    return OutContractException(msg)


def call_condf(cond_f: Callable[..., bool], f: Callable, result, args, kw: Dict[str, Any]) -> bool:
    try:
        callargs = list(getcallargs(f, *args, **kw).values())
        return cond_f(result, *callargs)
    except TypeError:
        ...

    f_sig = signature(f)
    f_varnames = list(f_sig.parameters)
    cond_sig = signature(cond_f)
    cond_varnames = list(cond_sig.parameters)
    varname_indexes = {v: f_varnames.index(v) for v in cond_varnames[1:]}

    def get_arg(v: str):
        if varname_indexes[v] < len(args):
            return args[varname_indexes[v]]
        elif v in kw:
            return kw[v]
        else:
            return f_sig.parameters[v].default

    var_args = {v: get_arg(v) for v in list(cond_sig.parameters.keys())[1:]}

    return cond_f(result, **var_args)


def out_contract(
        *conds,
        contract_tag: Optional[str] = None,
        contract_state: ContractState = DEFAULT_CONTRACT_STATE) -> Callable:

    def _out_contract(f: Callable) -> Callable:

        if contract_state.is_disable:
            return f

        @functools.wraps(f)
        def wrapped(*args, **kw) -> Callable:
            result = f(*args, **kw)
            if contract_state.is_disable:
                return f(*args, **kw)

            try:
                for cond_f in conds:
                    if not call_condf(cond_f, f, result, args, kw):
                        raise new_out_contract_exception(
                            f, list(args), kw, cond_f, result, contract_tag)
            except (ValueError, TypeError) as err:
                raise ContractError(*err.args)
            return result

        return wrapped

    return _out_contract



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ycontract']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'ycontract',
    'version': '0.3.1',
    'description': 'Python library for contracts testing',
    'long_description': 'ycontract\n================================================================================\n\n[![PyPI](https://img.shields.io/pypi/v/ycontract)](https://pypi.org/project/ycontract/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/ycontract)](https://pypi.org/project/ycontract/)\n[![pipeline status](https://gitlab.com/yassu/ycontract.py/badges/master/pipeline.svg)](https://gitlab.com/yassu/ycontract.py/-/pipelines/latest)\n[![coverage report](https://gitlab.com/yassu/ycontract.py/badges/master/coverage.svg)](https://gitlab.com/yassu/ycontract.py/-/commits/master)\n[![PyPI - License](https://img.shields.io/pypi/l/ycontract)](https://gitlab.com/yassu/ycontract.py/-/raw/master/LICENSE)\n\n\nPython library for contracts testing.\n\nThis library provides functions for checking argument(`in_contract`) and return value(`out_contract`) of a function.\n\nHow to install\n--------------------------------------------------------------------------------\n\n``` sh\n$ pip install ycontract\n```\n\nExample\n--------------------------------------------------------------------------------\n\nExample files are [here](https://gitlab.com/yassu/ycontract.py/-/blob/master/tests/test_contract.py)(test file)\n\nMain example is\n\n``` python\nfrom ycontract import contract, out_contract\n\n@contract(lambda a, b: a * b > 0)\ndef add(a, b, c):\n    return a + b\n\n\n@contract(returns=lambda res: res > 0)\ndef sub(a, b):\n    return a - b\n```\n\nAnd more complex example about in_contract is\n\n``` python\n@contract(\n    lambda a0: a0 > 0,\n    lambda a1, b: a1 > 0 and b > 0,\n    {\n        "a2": lambda x: x > 0,\n        "a3": [\n            lambda x: x >= 0,\n            lambda x: x < 4,\n        ],\n        ("a4", "a5"): lambda x, y: x > 0 and y > 0,\n    },\n    b=lambda x: x > 0,\n    contract_tag="tagtag",\n)\ndef add_for_complex(a0, a1, a2, a3, a4, a5, b=1):\n    return a0 + a1 + a2 + a3 + a4 + a5 + b\n```\n\nFurthermore if you want to be disable, call\n\n``` python\nycontract.disable_contract()\n```\n\nLICENSES\n--------------------------------------------------------------------------------\n\n[Apache 2.0](https://gitlab.com/yassu/ycontract.py/-/blob/master/LICENSE)\n',
    'author': 'yassu',
    'author_email': 'yasu0320.dev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/yassu/ycontract.py',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# flake8: noqa

__all__ = [
    # in ycontract._contract
    'contract',

    # in ycontract._in_contract
    'InContractException',
    'in_contract',

    # in ycontract._out_contract
    'OutContractException',
    'out_contract',

    # in ycontract.ycontract
    'DEFAULT_CONTRACT_STATE',
    'ContractError',
    'ContractException',
    'ContractState',
    'disable_contract',

    # in ycontract.util
    'exists_path',
    'in_range',
    'less_than',
    'less_than_or_eq',
    'more_than',
    'more_than_or_eq',
]

from ._contract import contract
from ._in_contract import InContractException, in_contract
from ._out_contract import OutContractException, out_contract
from .ycontract import (
    DEFAULT_CONTRACT_STATE,
    ContractError,
    ContractException,
    ContractState,
    disable_contract,
)

from .conds import exists_path, in_range, less_than, less_than_or_eq, more_than, more_than_or_eq  # isort:skip
